/*
  File: lib/acl_kernel.c
  (Linux Access Control List Management)

  Copyright (C) 1999 by Andreas Gruenbacher
  <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef __KERNEL__

#include <linux/fs.h>  /* also includes <linux/acl.h> */

#else

#include <errno.h>
#include <stdio.h>  /* for NULL */
#include <sys/types.h>
#include <sys/stat.h>  /* for S_IRWXU ... constants */
#include <sys/acl.h>

#include "acl_util.h"

#ifndef S_IRWXUGO
#  define S_IRWXUGO		(S_IRWXU|S_IRWXG|S_IRWXO)
#endif

#endif


/* acl_check_one()
 *
 * This function checks <aclcnt> ACL entries starting with <aclbufp>.
 * <type_flag> determines the type of ACL: 0 for an access ACL;
 * ACL_TYPE_DEFAULT for a default ACL.
 *
 * The ACL must be sorted.
 *
 * Return value:
 * -(ACL_*_ERROR) on error, a combination of ACL_HAS_* flags otherwise.
 */

#define FAIL_CHECK(error) \
	({ what = -error; goto getout; })

int acl_check_one(
	const acl_entry_t *aclbufp,
	int aclcnt,
	const acl_entry_t **last,
	int type_flag)
{
	const acl_entry_t *pe;
	int state = ACL_USER_OBJ;
	int what = 0;
	uid_t id = ACL_UNDEFINED_ID;
	int needs_class = 0;

	for (pe = aclbufp + aclcnt; aclbufp < pe; aclbufp++) {
		switch (aclbufp->a_type ^ type_flag) {
			/* case ACL_DEF_USER_OBJ: */
			case ACL_USER_OBJ:
				if (state == ACL_USER_OBJ) {
					id = ACL_UNDEFINED_ID;
					state = ACL_USER;
					break;
				}
				FAIL_CHECK(ACL_MULTI_ERROR);

			/* case ACL_DEF_USER: */
			case ACL_USER:
				if (state != ACL_USER)
					FAIL_CHECK(ACL_MISS_ERROR);
				if (id != ACL_UNDEFINED_ID &&
				    (aclbufp->a_id <= id ||
				     aclbufp->a_id == ACL_UNDEFINED_ID))
					FAIL_CHECK(ACL_DUPLICATE_ERROR);
				id = aclbufp->a_id;
				needs_class = 1;
				break;

			/* case ACL_DEF_GROUP_OBJ: */
			case ACL_GROUP_OBJ:
				if (state == ACL_USER) {
					id = ACL_UNDEFINED_ID;
					state = ACL_GROUP;
					break;
				}
				if (state >= ACL_GROUP)
					FAIL_CHECK(ACL_MULTI_ERROR);
				FAIL_CHECK(ACL_MISS_ERROR);

			/* case ACL_DEF_GROUP: */
			case ACL_GROUP:
				if (state != ACL_GROUP)
					FAIL_CHECK(ACL_MISS_ERROR);
				if (id != ACL_UNDEFINED_ID &&
				    (aclbufp->a_id <= id ||
				     aclbufp->a_id == ACL_UNDEFINED_ID)) {
					FAIL_CHECK(ACL_DUPLICATE_ERROR);
				}
				id = aclbufp->a_id;
				needs_class = 1;
				break;

			/* case ACL_DEF_MASK_OBJ: */
			case ACL_MASK_OBJ:
				if (state == ACL_GROUP) {
					state = ACL_OTHER_OBJ;
					break;
				}
				if (state >= ACL_OTHER_OBJ)
					FAIL_CHECK(ACL_MULTI_ERROR);
				FAIL_CHECK(ACL_MISS_ERROR);

			/* case ACL_DEF_OTHER_OBJ: */
			case ACL_OTHER_OBJ:
				if ((state == ACL_GROUP) && needs_class)
					FAIL_CHECK(ACL_MISS_ERROR);
				if ((state == ACL_GROUP) ||
				    (state == ACL_OTHER_OBJ)) {
				    	state = 0;
					break;
				}
				FAIL_CHECK(ACL_MISS_ERROR);

			default:
				FAIL_CHECK(ACL_ENTRY_ERROR);
		}
		if (aclbufp->a_perm & ~((mode_t)S_IRWXO)) {
			/* additional permission bit set */
			what |= ACL_HAS_BIT_ERROR;
		}
	}
	if (state != 0)
		FAIL_CHECK(ACL_MISS_ERROR);
	if (needs_class)
		what |= ACL_HAS_MASK;

	what = 0;

getout:

	if (last)
		*last = aclbufp;

	return what;
}


/*
  Returns 0 if the acl can be fully represented in the traditional permission
  bits of the mode field of a file. In that case, modep is set to that mode.
*/

int
acl_entries_equiv_mode(
	const acl_entry_t *aclentp,
	int aclcnt,
	mode_t *modep)
{
	const acl_entry_t *pe, *mask_obj = NULL;
	mode_t mode = 0;
	int found = 0, error = EINVAL;
	int not_equiv = 0;

	for (pe = aclentp + aclcnt; aclentp < pe; aclentp++) {
		switch (aclentp->a_type) {
			case ACL_USER_OBJ:
				if (found & ACL_USER_OBJ)
					goto fail;
				found |= ACL_USER_OBJ;
				mode |= (aclentp->a_perm << 6) & S_IRWXU;
				break;
			case ACL_GROUP_OBJ:
				if (found & ACL_GROUP_OBJ)
					goto fail;
				found |= ACL_GROUP_OBJ;
				mode |= (aclentp->a_perm << 3) & S_IRWXG;
				break;
			case ACL_OTHER_OBJ:
				if (found & ACL_OTHER_OBJ)
					goto fail;
				found |= ACL_OTHER_OBJ;
				mode |= (aclentp->a_perm << 0) & S_IRWXO;
				break;
			case ACL_MASK_OBJ:
				if (mask_obj)
					goto fail;
				mask_obj = aclentp;
				break;
			case ACL_USER:
			case ACL_GROUP:
				not_equiv = 1;
				break;
#ifndef __KERNEL__
			case ACL_DEF_USER_OBJ:
			case ACL_DEF_USER:
			case ACL_DEF_GROUP_OBJ:
			case ACL_DEF_GROUP:
			case ACL_DEF_MASK_OBJ:
			case ACL_DEF_OTHER_OBJ:
				error = ENOSYS;
				goto fail;
#endif
			default:
				goto fail;
		}
	}


        if (mask_obj && (((mode & S_IRWXG) >> 3) != mask_obj->a_perm)) {
                /* NOT POSIX DRAFT COMPLIANT !!! */
                mode &= (~S_IRWXG | (mask_obj->a_perm << 3));
                not_equiv = 1;
        }
        if (modep)
                *modep = (*modep & ~S_IRWXUGO) | (mode & S_IRWXUGO);

#ifdef __KERNEL__
        return not_equiv;
#else
	if (not_equiv) {
		error = ENOSYS;
		goto fail;
	}
	return 0;
#endif

fail:
#ifdef __KERNEL__
	return -error;
#else
	errno = error;
	return -1;
#endif
}


#ifdef __KERNEL__

int
acl_equiv_mode(
	const acl_head_t *acl,
	mode_t *modep)
{
	return acl_entries_equiv_mode(acl->h_ent, acl->h_entries, modep);
}

#else

int
acl_equiv_mode(
	const acl_t *aclp,
	mode_t *modep)
{
	return acl_entries_equiv_mode(aclp->entp, aclp->entries, modep);
}

#endif  /* __KERNEL__ */


void
mode_to_acl_entries(
	int mode,
	uid_t uid,
	gid_t gid,
	acl_entry_t *aclentp)
{
	aclentp[0].a_type = ACL_USER_OBJ;
	aclentp[0].a_id   = uid;
	aclentp[0].a_perm = (mode & S_IRWXU) >> 6;

	aclentp[1].a_type = ACL_GROUP_OBJ;
	aclentp[1].a_id   = gid;
	aclentp[1].a_perm = (mode & S_IRWXG) >> 3;

	//aclentp[2].a_type = ACL_MASK_OBJ;
	//aclentp[2].a_id   = ACL_UNDEFINED_ID;
	//aclentp[2].a_perm = (mode & S_IRWXG) >> 3;

	aclentp[2].a_type = ACL_OTHER_OBJ;
	aclentp[2].a_id   = ACL_UNDEFINED_ID;
	aclentp[2].a_perm = (mode & S_IRWXO);
}


#ifdef __KERNEL__

int
mode_to_acl(
	const struct inode *inode,
	acl_head_t **acl)
{
	if (!acl)
		return 3;

	*acl = acl_cache_alloc(3);
	if (!*acl)
		return -ENOMEM;

	mode_to_acl_entries(inode->i_mode, ACL_UNDEFINED_ID,
	                    ACL_UNDEFINED_ID, (*acl)->h_ent);
	return 3;
}

#else

int
mode_to_acl(
	const struct stat *buf,
	acl_t *aclp)
{
	if (!aclp)
		return 3;
	if (acl_reserve(aclp, 3) != 0)
		return -1;

	mode_to_acl_entries(buf->st_mode, buf->st_uid, buf->st_gid, aclp->entp);
	return 3;
}


#endif  /* __KERNEL__ */
