/*
  File: lib/file.c
  (Linux Access Control List Management)

  Last modified: Thu 23 Sep 1999, 00:08

  Copyright (C) 1999 by Andreas Gruenbacher
  <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <sys/acl.h>
#include "acl_util.h"


static
const char *
tabs(
	int max,
	const char *str);


int
acl_print(
	FILE *file,
	const acl_t *aclp,
	int what)
{
	mode_t acc_mask = S_IRWXO, def_mask = S_IRWXO, mask;
	int print_mask;
	acl_entry_t *p;
	int chars = 0, length, size = 256;
	char *textp;

	//if (acl_check(aclp, NULL) != 0)
	//	return -1;

	textp = (char*)malloc(size);
	if (textp == NULL)
		return -1;

	p = NULL;
	while ((p = acl_iterate((acl_t *)aclp, p))) {
		switch(p->a_type) {
			case ACL_MASK_OBJ:
				acc_mask = p->a_perm;
				break;

			case ACL_DEF_MASK_OBJ:
				def_mask = p->a_perm;
				break;
		}
	}
	p = NULL;
	while ((p = acl_iterate((acl_t *)aclp, p))) {
		do {
			length = acl_entry_to_str(p, textp, size);
			if (length == -1) {
				free(textp);
				return -1;
			}
			if (size < length) {
				while (size < length)
					size <<= 1;
				free(textp);
				textp = (char*)malloc(size);
				if (textp == NULL)
					return -1;
			}
		} while (size < length);

		mask = p->a_perm;
		switch(p->a_type) {
			case ACL_USER_OBJ:
			case ACL_MASK_OBJ:
			case ACL_OTHER_OBJ:
				if (what & ACL_PRINT_ACCESS_ACL)
					chars += fprintf(file, "%s\n", textp);
				continue;

			case ACL_DEF_USER_OBJ:
			case ACL_DEF_MASK_OBJ:
			case ACL_DEF_OTHER_OBJ:
				if (what & ACL_PRINT_DEFAULT_ACL)
					chars += fprintf(file, "%s\n", textp);
				continue;

			case ACL_USER:
			case ACL_GROUP_OBJ:
			case ACL_GROUP:
				if (!(what & ACL_PRINT_ACCESS_ACL))
					continue;
				mask = acc_mask;
				break;

			case ACL_DEF_USER:
			case ACL_DEF_GROUP_OBJ:
			case ACL_DEF_GROUP:
				if (!(what & ACL_PRINT_DEFAULT_ACL))
					continue;
				mask = def_mask;
				break;
		}

		print_mask  = (what & ACL_PRINT_ALL_EFFECTIVE) ||
		              ((what & ACL_PRINT_EFFECTIVE) &&
		                ((p->a_perm & mask) != p->a_perm));

		if (print_mask) {
			mask = p->a_perm & mask;
			chars += fprintf(file, "%s%s#effective:%c%c%c\n", textp,
				tabs(3, textp),
				(mask & S_IROTH) ? 'r' : '-',
				(mask & S_IWOTH) ? 'w' : '-',
				(mask & S_IXOTH) ? 'x' : '-');
		} else {
			chars += fprintf(file, "%s\n", textp);
		}
	}

	return (ferror(file) ? -1 : chars);
}


#define SKIP_WS(x) ({ \
	while (*(x) == ' ' || *(x) == '\t' || *(x) == '\n' || *(x) == '\r') \
		(x)++; \
	})

acl_t *
acl_read(
	FILE *file,
	int parse_mode,
	int *line,
	int *which,
	char **filenamep)
{
	char linebuf[1024];
	const char *c;
	acl_t *aclp = NULL;
	acl_entry_t *aclentp;
	uid_t uid = ACL_UNDEFINED_ID;
	gid_t gid = ACL_UNDEFINED_ID;

	if (line)
		*line = 0;
	if (which)
		*which = 0;
	if (filenamep)
		*filenamep = NULL;

	for(;;) {
		if (line) (*line)++;
		if (fgets(linebuf, sizeof(linebuf), file) == NULL)
			break;

		c = linebuf;
		SKIP_WS(c);
		if (*c == '\0') {
			if (parse_mode & ACL_PARSE_MULTI)
				break;
			else
				continue;
		}

		if (aclp == NULL) {
			aclp = acl_init(0);
			if (aclp == NULL)
				goto fail;
		}

		if (*c == '#') {
			/*
			  Only scan comments before the ACL
			  entries for `file:', `owner:' and `group:'
			*/
			if (aclp->entries != 0)
				continue;

			c++;
			SKIP_WS(c);
			if (strncmp(c, "file:", 5) == 0) {
				if (filenamep && *filenamep == NULL) {
					c += 5;
					SKIP_WS(c);
					
					*filenamep = (char*)malloc(strlen(c)+1);
					if (!*filenamep) {
						errno = ENOMEM;
						goto fail;
					}
					strcpy(*filenamep, c);
					c = strrchr(*filenamep, '\0');
					while (c > *filenamep &&
					                (*(c-1) == '\r' ||
					                 *(c-1) == '\n')) {
						*(char *)(--c) = '\0';
					}
				}
			} else if (strncmp(c, "owner:", 6) == 0) {
				if (uid == ACL_UNDEFINED_ID) {
					c += 6;
					SKIP_WS(c);
					
					if (parse_user(&c, &uid) == 0) {
						SKIP_WS(c);
						if (*c != '\0')
							 uid = ACL_UNDEFINED_ID;
					}
				}
			} else if (strncmp(c, "group:", 6) == 0) {
				if (gid == ACL_UNDEFINED_ID) {
					c += 6;
					SKIP_WS(c);
					
					if (parse_group(&c, &gid) == 0) {
						SKIP_WS(c);
						if (*c != '\0')
							 gid = ACL_UNDEFINED_ID;
					}
				}
			}
			continue;
		}

		if (acl_create_entry(aclp, &aclentp) != 0)
			goto fail;

		if (parse_aclent(linebuf, &c, aclentp, ACL_PARSE_PERM) != 0) {
			if (which)
				*which = (c - linebuf);
			errno = EINVAL;
			goto fail;
		}

		if (aclentp->a_type == ACL_USER_OBJ &&
		    uid != ACL_UNDEFINED_ID) {
		    	aclentp->a_id = uid;
		} else if (aclentp->a_type == ACL_GROUP_OBJ &&
		           gid != ACL_UNDEFINED_ID) {
			aclentp->a_id = gid;
		}

		SKIP_WS(c);
		if (*c != '\0' && *c != '#') {
			if (which)
				*which = (c - linebuf);
			errno = EINVAL;
			goto fail;
		}
	}

	if (ferror(file))
		goto fail;
	errno = 0;
	return aclp;

fail:
	if (aclp != NULL)
		acl_free(aclp);
	return NULL;
}


#define TABS "\t\t\t\t\t\t\t\t"
static const char *tabs(int max, const char *str) {
	static const char *tabs = TABS;
	const char *tabse = tabs + sizeof(TABS) - 1;
	int n = strlen(str)/8;
	if (max > sizeof(TABS)-1) max = sizeof(TABS)-1;
	if (n >= max)
		return tabse-1;
	else
		return tabse-max+n-1;
}
#undef TABS

