/*
  File: lib/libcall.c
  (Linux Access Control List Management)

  Last modified: Wed 22 Sep 1999, 23:50

  Copyright (C) 1999 by Andreas Gruenbacher
  <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <sys/acl.h>
#include "acl_util.h"


/* helper functions */
int
setacl_noacl(
	const char *pathp,
	const struct stat *buf,
	const acl_entry_t *aclbuf,
	int aclcnt);
int
setfacl_noacl(
	int filedes,
	const struct stat *buf,
	const acl_entry_t *aclbufp,
	int aclcnt);


/* the actual system calls */
int
acl_sorted(
	const char *pathp,
	int cmd,
	int aclcnt,
	acl_entry_t *aclbufp);
int
facl_sorted(
	int filedes,
	int cmd,
	int aclcnt,
	acl_entry_t *aclbufp);


int
acl(
	const char *pathp,
	int cmd,
	int aclcnt,
	acl_entry_t *aclbufp)
{
	struct stat buf;
	int result;

#ifdef USE_ACL_SYSCALLS
	if (cmd == ACL_SETACL)
		acl_entry_sort(aclbufp, aclcnt);
	result = acl_sorted(pathp, cmd, aclcnt, aclbufp);
#else
	result = EOF;
	errno = ENOSYS;
#endif

	if (result == EOF && errno == ENOSYS) {
		/* Not implemented -- try to use permission bits */

		/* should we do a stat or lstat here ? */
		if (stat(pathp, &buf) != 0)
			return EOF;
		
		switch (cmd) {
			case ACL_GETACL:
				if (aclcnt < 3) {
					errno = ENOSPC;
					return -1;
				}
				mode_to_acl_entries(buf.st_mode, buf.st_uid, buf.st_gid, aclbufp);
				return 3;

			case ACL_SETACL:
				return setacl_noacl(pathp, &buf,
				                    aclbufp, aclcnt);

			case ACL_GETACLCNT:
				return 3;  /* number of entries */

			default:
				errno = EINVAL;
				return EOF;
		}
	} else {
		return result;
	}
}


int
facl (
	int filedes,
	int cmd,
	int aclcnt,
	acl_entry_t *aclbufp)
{
	int result;
	struct stat buf;

#ifdef USE_ACL_SYSCALLS
	if (cmd == ACL_SETACL)
		acl_entry_sort(aclbufp, aclcnt);
	result = facl_sorted(filedes, cmd, aclcnt, aclbufp);
#else
	result = EOF;
	errno = ENOSYS;
#endif

	if (errno == ENOSYS) {
		/* Not implemented -- try to use permission bits */
		
		if (fstat(filedes, &buf) != 0)
			return EOF;

		switch (cmd) {
			case ACL_GETACL:
				if (aclcnt < 3) {
					errno = ENOSPC;
					return -1;
				}
				mode_to_acl_entries(buf.st_mode, buf.st_uid, buf.st_gid, aclbufp);
				return 3;

			case ACL_SETACL:
				return setfacl_noacl(filedes, &buf,
				                     aclbufp, aclcnt);

			case ACL_GETACLCNT:
				return 3;  /* number of entries */

			default:
				errno = EINVAL;
				return EOF;
		}
	} else {
		return result;
	}
}


#define ALL_PERM (S_IRWXU | S_IRWXG | S_IRWXO)

int
setacl_noacl(
	const char *pathp,
	const struct stat *buf,
	const acl_entry_t *aclbufp,
	int aclcnt)
{
	mode_t mode;

	if (acl_entries_equiv_mode(aclbufp, aclcnt, &mode) != 0) {
		errno = ENOSYS;
		return EOF;
	}

	/*
	  map in all other bits from struct stat
	  (S_ISUID, S_ISGID, S_ISVTX, ...)
	*/
	mode = (mode & ALL_PERM) | (buf->st_mode & ~ALL_PERM);

	return chmod(pathp, mode);
}


int
setfacl_noacl(
	int filedes,
	const struct stat *buf,
	const acl_entry_t *aclbufp,
	int aclcnt)
{
	mode_t mode;

	if (acl_entries_equiv_mode(aclbufp, aclcnt, &mode) != 0) {
		errno = ENOSYS;
		return EOF;
	}

	/*
	  map in all other bits from struct stat
	    (S_ISUID, S_ISGID, ...)
	*/
	mode = (mode & ALL_PERM) | (buf->st_mode & ~ALL_PERM);

	return fchmod(filedes, mode);
}

#undef ALL_PERM

