/*
  File: lib/manip.c
  (Linux Access Control List Management)

  Last modified: Thu 23 Sep 1999, 16:28

  Copyright (C) 1999 by Andreas Gruenbacher
  <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/acl.h>
#include "acl_util.h"


acl_t *
acl_init(
	int entries)
{
	acl_t *aclp;

	if (entries < 0) {
		errno = EINVAL;
		return NULL;
	}

	aclp = (acl_t *)malloc(sizeof(acl_t));
	if (aclp == NULL) {
		errno = ENOMEM;
		return NULL;
	}
	aclp->size = entries;
	aclp->entries = 0;
	if (entries > 0) {
		aclp->entp = (acl_entry_t *)malloc(entries *
		                                   sizeof(acl_entry_t));
		if (aclp->entp == NULL) {
			errno = ENOMEM;
			free(aclp);
			return NULL;
		}
	} else {
		aclp->entp = NULL;
	}
	return aclp;
}


int
acl_reserve(
	acl_t *aclp,
	int entries)
{
	int size;
	acl_entry_t *ent2p;

	if (entries < 0) {
		errno = EINVAL;
		return -1;
	}

	entries += aclp->entries;
	if (aclp->size < entries) {
		size = (aclp->size > 0) ? aclp->size : 1;
		while (size < entries)
			size <<= 1;

		if (aclp->entp) {
			ent2p = (acl_entry_t *)realloc(aclp->entp, size *
			                               sizeof(acl_entry_t));
		} else {
			ent2p = (acl_entry_t *)malloc(size *
			                              sizeof(acl_entry_t));
		}

		if (ent2p != NULL) {
			aclp->entp = ent2p;
			aclp->size = size;
		} else {
			errno = ENOMEM;
			return -1;
		}
	}
	return 0;
}


acl_t *
acl_dup(
	const acl_t *aclp)
{
	acl_t *dupp;

	dupp = (acl_t *)malloc(sizeof(acl_t));
	if (dupp == NULL) {
		errno = ENOMEM;
		return NULL;
	}
	dupp->size = dupp->entries = aclp->entries;
	dupp->entp = (acl_entry_t *)malloc(aclp->entries * sizeof(acl_entry_t));
	if (dupp->entp == NULL) {
		errno = ENOMEM;
		free(dupp);
		return NULL;
	}
	memcpy(dupp->entp, aclp->entp, aclp->entries * sizeof(acl_entry_t));
	return dupp;
}


int
acl_free(
	acl_t *aclp)
{
	free(aclp->entp);
	free(aclp);
	return 0;
}


int
acl_create_entry(
	acl_t *aclp,
	acl_entry_t **entp)
{
	/* make sure there is room for another entry */
	if (acl_reserve(aclp, 1) != 0)
		return -1;

	*entp = &aclp->entp[aclp->entries++];
	(*entp)->a_type = ACL_UNDEFINED_TAG;
	(*entp)->a_id   = ACL_UNDEFINED_ID;
	(*entp)->a_perm = 0;
	return 0;
}


int
acl_delete_entry(
	acl_t *aclp,
	acl_entry_t *entp)
{
	if (entp < aclp->entp ||
	    entp >= (aclp->entp + aclp->entries)) {
		errno = EINVAL;
		return -1;
	}

	if (entp < aclp->entp + aclp->entries-1) {
		memmove(entp, entp+1, (aclp->entries - (entp-aclp->entp) - 1) *
		                      sizeof(acl_entry_t));
	}
	aclp->entries--;
	/* If (entries < size/4) realloc size/2 */
	return 0;
}


int
acl_copy_entry(
	acl_entry_t *destp,
	const acl_entry_t *srcp)
{
	*destp = *srcp;
	return 0;
}


acl_entry_t *
acl_iterate(
	acl_t *aclp,
	acl_entry_t *entp)
{
	if (entp == NULL) {
		entp = aclp->entp;
	} else {
		entp++;
	}
	if (entp < aclp->entp + aclp->entries)
		return entp;
	else
		return NULL;
}


typedef int (*compare)(const void *, const void *);
static int
acl_entry_compare(
	acl_entry_t *a,
	acl_entry_t *b)
{
        if (a->a_type < b->a_type)
                return -1;
        else if (a->a_type > b->a_type)
                return 1;

        if (a->a_id < b->a_id)
                return -1;
        else if (a->a_id > b->a_id)
                return 1;
        else
                return 0;
}


void
acl_entry_sort(
	acl_entry_t *aclentp,
	int aclcnt)
{
	qsort(aclentp, aclcnt, sizeof(acl_entry_t),
              (compare)acl_entry_compare);
}


int
acl_sort(
	acl_t *aclp)
{
	acl_entry_sort(aclp->entp, aclp->entries);
	return 0;
}


int
acl_check(
	acl_t *aclp,
	int *whichp)
{
	int acltype = 0, acltype2;
	acl_entry_t *backup, *lastp, *endp;

	if (aclp->entries == 0) {
		if (whichp)
			 *whichp = 0;
		return 0;
	}

	/* canonical sort order */
	if (acl_sort(aclp) != 0)
		return -1;

	/* access ACL */
	backup = aclp->entp;
	endp = aclp->entp + aclp->entries;
	if (backup != endp && !(backup->a_type & ACL_TYPE_DEFAULT)) {
		lastp = backup;
		while (lastp < endp && !(lastp->a_type & ACL_TYPE_DEFAULT))
			lastp++;

		acltype = acl_check_one(backup, lastp - backup,
		                        (const acl_entry_t **)&backup,
		                        0);
		if (whichp)
			*whichp = (backup - aclp->entp + 1);
		if (acltype < 0)
			return acltype;

		acltype |= ACL_HAS_ACCESS_ACL;
	}
	
	/* default ACL */
	if (backup != endp && (backup->a_type & ACL_TYPE_DEFAULT)) {
		acltype2 = acl_check_one(backup, endp - backup,
		                         (const acl_entry_t **)&backup,
		                         ACL_TYPE_DEFAULT);
		if (whichp)
			*whichp = (backup - aclp->entp + 1);
		if (acltype2 < 0)
			return acltype2;

		acltype |= ACL_HAS_DEFAULT_ACL;
		acltype |= (acltype2 & ACL_HAS_MASK) ? ACL_HAS_DEFAULT_MASK : 0;
		acltype |= (acltype2 & ACL_HAS_BIT_ERROR);
	}

	return acltype;
}


const char *
acl_strerror(
	int code)
{
	switch(code) {
		case ACL_MULTI_ERROR:
			return "multiple entries";
		case ACL_DUPLICATE_ERROR:
			return "duplicate entries";
		case ACL_MISS_ERROR:
			return "missing or wrong entry";
		case ACL_ENTRY_ERROR:
			return "invalid entry type";
		default:
			return NULL;
	}
}


char *
acl_to_text(
	const acl_t *aclp)
{
	size_t size = (aclp->entries * 20) + 1;  /* guess... */
	char *textp, *tmp;
	acl_entry_t *aclentp;
	size_t index=0, length;

	textp = (char*)malloc(size);
	if (textp == NULL) return NULL;

	aclentp = acl_iterate((acl_t *)aclp, NULL);
	while (aclentp) {
		length = acl_entry_to_str(aclentp, textp + index, size - index);
		if (length == -1) {
			free(textp);
			return NULL;
		} else {
			length++;  /* we need one more character */
		}

		if (index + length >= size) {
			/* enlarge buffer */
			if (length > size)
				size += length;
			else
				size *= 2;
			tmp = (char*)realloc(textp, size);
			if (tmp == NULL) {
				free(textp);
				return NULL;
			}
			textp = tmp;
			/* retry this entry */
		} else {
			index += length;
			aclentp = acl_iterate((acl_t *)aclp, aclentp);
			if (aclentp)
				textp[index-1] = ',';
		}
	}
	textp[index] = '\0';
	return textp;
}


acl_t *
acl_from_text(
	const char *textp)
{
	return parse_acl(textp, NULL, ACL_PARSE_WITH_PERM |
	                              ACL_PARSE_NO_RELATIVE);
}


int guess_size = 12;

acl_t *
acl_get_file(
	const char *pathp,
	int type)
{
	int error;
	acl_t *aclp;
	acl_entry_t *aclentp, *pe;

	aclp = acl_init(guess_size);
	if (aclp == NULL)
		return NULL;

	error = acl(pathp, ACL_GETACL, aclp->size, aclp->entp);
	if (error <= 0) {
		if (errno != ENOMEM) {
			goto fail;
		} else {
			error = acl(pathp, ACL_GETACLCNT, 0, NULL);
			if (error < 0)
				goto fail;
			if (acl_reserve(aclp, error) != 0)
				goto fail;
			error = acl(pathp, ACL_GETACL,
			            aclp->size, aclp->entp);
			if (error < 0)
				goto fail;
			aclp->entries = error;

                        /*
                                We succeeded by enlarging the buffer. Next
				time, we try a bigger buffer size.
                        */
                        guess_size <<= 1;
		}
	} else {
		aclp->entries = error;
	}

	if (type == 0)
		type = ACL_TYPE_ACCESS | ACL_TYPE_DEFAULT;

	if (!(type & ACL_TYPE_ACCESS)) {
		/* Remove access acl */
		aclentp = aclp->entp;
		pe = aclp->entp + aclp->entries;
		while ((aclentp < pe) && !(aclentp->a_type &
		                           ACL_TYPE_DEFAULT)) {
			aclentp++;
		}
		aclp->entries = (pe - aclentp);
		memmove(aclp->entp, aclentp, aclp->entries *
		                             sizeof(acl_entry_t));

		/* Remove default acl bits */
		aclentp = aclp->entp;
		pe = aclp->entp + aclp->entries;
		while (aclentp < pe) {
			aclentp->a_type &= ~ACL_TYPE_DEFAULT;
			aclentp++;
		}
	}

	if (!(type & ACL_TYPE_DEFAULT)) {
		/* Remove default acl */
		aclentp = aclp->entp;
		pe = aclp->entp + aclp->entries;
		while ((pe > aclentp) && ((pe-1)->a_type &
		                           ACL_TYPE_DEFAULT)) {
			pe--;
		}
		aclp->entries = (pe - aclentp);
	}

	return aclp;

fail:
	acl_free(aclp);
	return NULL;
}


int acl_find(acl_t *aclp,
             acl_entry_t **user_obj,
	     acl_entry_t **group_obj,
             acl_entry_t **mask_obj,
	     acl_entry_t **other_obj)
{
  acl_entry_t *pa, *pe;
  if (user_obj)  *user_obj = NULL;
  if (group_obj) *group_obj = NULL;
  if (mask_obj)  *mask_obj = NULL;
  if (other_obj) *other_obj = NULL;

  for (pa = aclp->entp, pe = pa + aclp->entries;
       pa != pe;
       pa++) {
    switch (pa->a_type) {
      case ACL_USER_OBJ:
        if (user_obj)
          (*user_obj) = pa;
        break;
      case ACL_GROUP_OBJ:
        if (group_obj)
          *group_obj = pa;
        break;
      case ACL_MASK_OBJ:
        if (mask_obj)
          *mask_obj = pa;
        break;
      case ACL_OTHER_OBJ:
        if (other_obj)
          *other_obj = pa;
        break;
    }
  }
        return 0;
}

