/*
  File: lib/text.c
  (Linux Access Control List Management)

  Last modified: Wed 22 Sep 1999, 23:40

  Copyright (C) 1999 by Andreas Gruenbacher
  <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/acl.h>
#include "acl_util.h"


#define SKIP_WS(x) \
	({ while (*(x) == ' ' || *(x) == '\t') (x)++; })


/*
	Parses the next acl entry in textp.
	Stores the acl entry in aclent.

	Returns:
		-1 on error, and 0 otherwise.

	** does not set errno **
*/

int
parse_aclent(
	const char *textp,
	const char **endp,
	acl_entry_t *aclent,
	int parse_mode)
{
	const char *backup;
	int perm_chars;

	/* clear aclent */
	aclent->a_type = 0;
	aclent->a_id = ACL_UNDEFINED_ID;
	aclent->a_perm = 0;

	*endp = textp;

	/* check for default acl entry */
	SKIP_WS(*endp);
	if (**endp == 'd') {
		backup = *endp;

		if (strncmp(*endp, "default", 7) == 0)
			*endp += 7;
		else
			(*endp)++;
		SKIP_WS(*endp);
		if (**endp == ':') {
			(*endp)++;
			aclent->a_type |= ACL_TYPE_DEFAULT;
		} else {
			*endp = backup;
		}
	}

	/* parse acl entry type */
	SKIP_WS(*endp);
	switch (**endp) {
		case 'u':  /* user */
			backup = *endp;

			if (strncmp(*endp, "user", 4) == 0)
				*endp += 4;
			else
				(*endp)++;

			SKIP_WS(*endp);
			if (**endp == ':') {
				(*endp)++;
				SKIP_WS(*endp);

user_entry:
				if (parse_user(endp, &aclent->a_id) == 0) {
					/* (default?) owner */
					aclent->a_type |= ACL_USER;
				} else {
					/* (default?) additional user */
					aclent->a_type |= ACL_USER_OBJ;
				}
			} else {
				/*
				if (!(parse_mode & ACL_PARSE_WITH_PERM))
					aclent->a_type |= ACL_USER_OBJ;
				else
					return -1;
				*/
				*endp = backup;
				goto user_entry;
			}
			break;

		case 'g':  /* group */
			backup = *endp;

			if (strncmp(*endp, "group", 5) == 0)
				*endp += 5;
			else
				(*endp)++;

			SKIP_WS(*endp);
			if (**endp == ':') {
				(*endp)++;
				SKIP_WS(*endp);

				if (parse_group(endp, &aclent->a_id) == 0) {
					/* (default?) owning group */
					aclent->a_type |= ACL_GROUP;
				} else {
					/* (default?) additional group */
					aclent->a_type |= ACL_GROUP_OBJ;
				}
			} else {
				/*
				if (!(parse_mode & ACL_PARSE_WITH_PERM))
					aclent->a_type |= ACL_GROUP_OBJ;
				else
					return -1;
				*/
				*endp = backup;
				goto user_entry;
			}
			break;

		case 'o':  /* other */
			backup = *endp;

			if (strncmp(*endp, "other", 5) == 0)
				*endp += 5;
			else
				(*endp)++;

			/* test includes '\0' */
			if (strchr(":, \t", **endp) != NULL) {
				/* (default?) other users */
				aclent->a_type |= ACL_OTHER_OBJ;
			} else {
				*endp = backup;
				goto user_entry;
			}
			break;

		case 'm':  /* mask */
			backup = *endp;

			if (strncmp(*endp, "mask", 4) == 0)
				*endp += 4;
			else
				(*endp)++;
			
			/* test includes '\0' */
			if (strchr(":, \t", **endp) != NULL) {
				/* (default?) acl mask */
				aclent->a_type |= ACL_MASK_OBJ;
			} else {
				*endp = backup;
				goto user_entry;
			}
			break;

		default:  /* assume "user:" */
			goto user_entry;
			//return -1;
	}

	SKIP_WS(*endp);
	if (*(*endp) != ':') {
		if (!(parse_mode & ACL_PARSE_NO_PERM))
			return -1;
		return 0;
	}
	(*endp)++;

	if (!(parse_mode & ACL_PARSE_WITH_PERM))
		return 0;

	SKIP_WS(*endp);
	if (parse_mode & ACL_PARSE_WITH_RELATIVE) {
		if (**endp == '+') {
			(*endp)++;
			aclent->a_type |= ACL_SEQ_ADD;
		} else if (**endp == '^') {
			(*endp)++;
			aclent->a_type |= ACL_SEQ_SUBTRACT;
		} else {
			if (!(parse_mode & ACL_PARSE_NO_RELATIVE))
				return -1;

			aclent->a_type |= ACL_SEQ_REPLACE;
		}
	}

	/* parse permissions */
	SKIP_WS(*endp);
	if (*(*endp) >= '0' && *(*endp) <= '7') {
		aclent->a_perm = 0;
		while (*(*endp) == '0') (*endp)++;
		if (*(*endp) >= '1' && *(*endp) <= '7') {
			aclent->a_perm = (*(*endp)++ - '0');
		}

		return 0;
	}

	for (perm_chars=0; perm_chars<3; perm_chars++, (*endp)++) {
		switch(**endp) {
			case 'r':
				if (aclent->a_perm & S_IROTH) return -1;
				aclent->a_perm |= S_IROTH;
				break;

			case 'w':
				if (aclent->a_perm & S_IWOTH) return -1;
				aclent->a_perm |= S_IWOTH;
				break;

			case 'x':
				if (aclent->a_perm & S_IXOTH) return -1;
				aclent->a_perm |= S_IXOTH;
				break;

			case '-':
				/* ignore */
				break;

			default:
				return (perm_chars > 0) ? 0 : -1;
		}
	}

	return (perm_chars == 3) ? 0 : -1;
}


/*
	Parse a comma-separated list of acl entries.

	which is set to the first character that was not parsed. This is
	      the end of textp in case of success.
	aclcnt is set to the number of acl entries in the return value.
	perm determines whether the acl entries have permissions:
	     Set to ACL_PARSE_PERM or ACL_PARSE_NOPERM.

	Upon success, returns a new acl allocated with malloc(). The user
	is responsible to free() the buffer when no longer needed.
	Upon failure NULL is returned, and errno is set to a meaningful value.
*/
acl_t *
parse_acl(
	const char *textp,
	const char **which,
	int parse_mode)
{
	int count;
	const char *pe = textp;
	acl_t *aclp = NULL;
	acl_entry_t *aclentp;

	if (textp == NULL) {
		errno = EFAULT;
		goto cleanup;
	}

	aclp = acl_init(0);
	if (!aclp) {
		errno = ENOMEM;
		goto cleanup;
	}

	for (count=0; *textp != '\0'; count++, textp = pe) {
		if (acl_create_entry(aclp, &aclentp) != 0)
			goto cleanup;
		if (parse_aclent(textp, &pe, aclentp, parse_mode) != 0) {
			errno = EINVAL;
			goto cleanup;
		}
		SKIP_WS(pe);
		if (*pe != ',')
			break;
		pe++;
	}

	if (*pe != '\0') {
		errno = EINVAL;
		goto cleanup;
	}

cleanup:

	if (errno && aclp) {
		acl_free(aclp);
		aclp = NULL;
	}
	if (which) *which = pe;

	return aclp;
}


int
id_tostr(
	char **textp,
	size_t size,
	uid_t uid)
{
	uid_t u = uid;
	int digits = 1;
	uid_t factor = 1;

	while ((u /= 10) != 0) {
		digits++;
		factor *= 10;
	}
	if (size && (uid == 0)) {
		*(*textp)++ = '0';
	} else {
		while (size && factor>0) {
			*(*textp)++ = '0' + (uid / factor);
			size--;
			uid %= factor;
			factor /= 10;
		}
	}
	if (size) {
		*(*textp) = '\0';
		size--;
	}

	return digits;
}


#define ADVANCE(x) \
	sz += (x); \
	if (size > (x)) { \
		textp += (x); \
		size -= (x); \
	} else { \
		textp += size; \
		size = 1; \
	}

/*
	returns the size of the buffer that was/would be filled.
*/

int
acl_entry_to_str(
	const acl_entry_t *aclent,
	char *textp,
	size_t size)
{
	size_t sz = 0;
	int type;
	const char *str;

	if (aclent == NULL)
		return -1;

	if (textp == NULL)
		size = 0;

	if (aclent->a_type & ACL_TYPE_DEFAULT) {
		strncpy(textp, "default:", size);
		ADVANCE(8);
	}

	type = aclent->a_type & ~ACL_TYPE_DEFAULT & ~ACL_SEQ_CMD;
	switch (type) {
		case ACL_USER_OBJ:  /* owner */
		case ACL_USER:  /* additional user */
			strncpy(textp, "user:", size);
			ADVANCE(5);
			if (type == ACL_USER &&
			    (aclent->a_id != ACL_UNDEFINED_ID)) {
				str = user_name(aclent->a_id);
				if (str != NULL) {
					strncpy(textp, str, size);
					ADVANCE(strlen(str));
				} else {
					sz += id_tostr(&textp, size,
					               aclent->a_id);
				}
			}
			*textp = ':';
			ADVANCE(1);
			break;

		case ACL_GROUP_OBJ:  /* owning group */
		case ACL_GROUP:  /* additional group */
			strncpy(textp, "group:", size);
			ADVANCE(6);
			if (type == ACL_GROUP &&
			    (aclent->a_id != ACL_UNDEFINED_ID)) {
				str = group_name(aclent->a_id);
				if (str != NULL) {
					strncpy(textp, str, size);
					ADVANCE(strlen(str));
				} else {
					sz += id_tostr(&textp, size,
					               aclent->a_id);
				}
			}
			*textp = ':';
			ADVANCE(1);
			break;

			
		case ACL_MASK_OBJ:  /* acl mask */
			strncpy(textp, "mask:", size);
			ADVANCE(5);
			break;

		case ACL_OTHER_OBJ:  /* other users */
			strncpy(textp, "other:", size);
			ADVANCE(6);
			break;

		default:
			return 0;
	}

	type = aclent->a_type & ACL_SEQ_CMD;
	switch(type) {
		case ACL_SEQ_ADD:
			*textp = '+';
			ADVANCE(1);
			break;
		case ACL_SEQ_SUBTRACT:
			*textp = '^';
			ADVANCE(1);
			break;
		default:
			break;
	}

	switch ((size > 3) ? 4 : size) {
		case 4:
			textp[2] = (aclent->a_perm & S_IXOTH) ? 'x' : '-'; 
			/* fall through */
		case 3:
			textp[1] = (aclent->a_perm & S_IWOTH) ? 'w' : '-'; 
			/* fall through */
		case 2:
			textp[0] = (aclent->a_perm & S_IROTH) ? 'r' : '-'; 
			/* fall through */
		default:
			/* nothing */
	}
	ADVANCE(3);

	/* zero-terminate string */
	if (size > 0) {
		*textp++ = '\0';
		size--;
	}
	
	return sz;  /* total size required -- may be more that size was */
}

#undef ADVANCE



