/*
  File: progs/getfacl.c
  (Linux Access Control List Management)

  Last modified: Wed 22 Sep 1999, 23:42

  Copyright (C) 1999 by Andreas Gruenbacher
  <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/acl.h>
#include <dirent.h>
#include <getopt.h>
#include "acl_util.h"

#define VERSION "0.5.0"

#define CMD_LINE_OPTIONS "adceEsRY5Pvh"
#define CMD_LINE_SPEC "[-" CMD_LINE_OPTIONS "] file ..."

#define POSIXLY_CORRECT "POSIXLY_CORRECT"


const char *progname;
int opt_acl_only = 0;
int opt_recurse = 0;  /* recurse */
int opt_follow_symlinks = 0;  /* follow symlinks */
int opt_no_comments = 0;  /* suppress comments */
int opt_postorder = 0;  /* postorder traversal? (quite weird for a read...) */
int opt_strip_leading_slash = 1;
int posixly_correct = 0;  /* Posix compatible behavior? */
int had_errors = 0, had_warnings = 0;
int absolute_warning = 0;


int do_print(const char *pathp, int what);
int do_get(const char *pathp, int what, int toplevel);


/*
 * Get the ACL of the file/directory <pathp>.
 * <toplevel> is true at the topmost level only, so that symlinks
 * are followed if a symlink is given as a command-line argument,
 * despite their being ignored on all levels below.
 * <what> is a combination of the ACL_PRINT_* constants defined in
 * acl_util.h
 */

int do_get(const char *pathp, int what, int toplevel)
{
	struct stat buf;
	DIR *dir;
	struct dirent *dirent;
	char *ipathp = (char *)pathp;  /* path including trailing '/' */
	char *subpathp = NULL; /* path with subdirectory appended */
	char *lastslash;
	int recurse = 0;  /* recurse into this directory / follow symlink? */
	int skip = 0;  /* skip this entry? */

	if (lstat(pathp, &buf) != 0)  /* lstat: stat links, don't follow */
		goto fail;
	
	lastslash = strrchr(pathp, '/');

	switch (buf.st_mode & S_IFMT) {
		case S_IFLNK:  /* symbolic link (``@'') */
			skip = (!toplevel || opt_follow_symlinks);
			/* fall through */

		case S_IFDIR:  /* directory (``/'') */
			/* make sure it ends in a slash... */
			if (!lastslash || *(lastslash+1) != '\0') {
				ipathp = (char*)malloc(strlen(pathp)+2);
				if (ipathp == NULL)
					{ errno = ENOMEM; goto fail; }
				strcpy(ipathp, pathp);
				strcat(ipathp, "/");
			}
				
			if ((buf.st_mode & S_IFMT) == S_IFDIR)
				recurse = opt_recurse;
			else
				recurse = (opt_recurse && opt_follow_symlinks);
			break;

		case S_IFREG:  /* regular file */
		case S_IFBLK:  /* block device */
		case S_IFCHR:  /* character device */
		case S_IFIFO:  /* named pipe (``|'') */
		case S_IFSOCK:  /* socket (``='') */
			/* make sure it doesn't end in a slash... */
			if (lastslash && (*lastslash+1) == '\0') {
				ipathp = (char*)malloc(strlen(pathp)+1);
				if (ipathp == NULL)
					{ errno = ENOMEM; goto fail; }
				strcpy(ipathp, pathp);
				lastslash = strrchr(ipathp, '/');
				assert (lastslash != NULL);
				*lastslash = '\0';
			}
			break;

		default:
			errno = EINVAL;
			goto fail;
	}

	if (!skip && !opt_postorder) {
		if (do_print(/*ipathp*/ pathp, what) == -1) {
			if (recurse && errno == EACCES) {
				fprintf(stderr, "%s: %s: %s\n", progname,
				        pathp, strerror(errno));
				had_errors++;
				return 0;
			} else {
				/* error is handled outside of this function */
				goto fail;
			}
		}
	}

	if (recurse) {
		dir = opendir(ipathp);
		if (dir == NULL) {
			if (errno == EACCES) {
				fprintf(stderr, "%s: %s\n",
				        pathp, strerror(errno));
				had_errors++;
				return 0;
			} else {
				/* error is handled outside of this function */
				goto fail;
			}
		}
		while ((dirent = readdir(dir)) != NULL) {
			if (*dirent->d_name == '.') {
				if (*(dirent->d_name+1) == '\0' ||
				    (*(dirent->d_name+1) == '.' &&
				     *(dirent->d_name+2) == '\0'))
					continue;
			}

			subpathp = (char*)malloc(strlen(ipathp) +
					    strlen(dirent->d_name) + 1);
			if (subpathp == NULL)
				goto fail;

			strcpy(subpathp, ipathp);
			strcat(subpathp, dirent->d_name);
			do_get(subpathp, what, 0);  /* recurse */
			free(subpathp);
		}
		closedir(dir);
	}

	if (!skip && opt_postorder) {
		if (do_print(/*ipathp*/ pathp, what) == -1) {
			if (recurse && errno == EACCES) {
				fprintf(stderr, "%s: %s: %s\n", progname,
				        pathp, strerror(errno));
				had_errors++;
				return 0;
			} else {
				/* error is handled outside of this function */
				goto fail;
			}
		}
	}

	if (ipathp != pathp)
		free(ipathp);
	return 0;

fail:
	if (ipathp != pathp)
		free(ipathp);
	return -1;
}


int do_print(const char *pathp, int what)
{
	acl_entry_t *user_obj, *group_obj;
	const char *owner, *group;
	const char *prefix = "";
	//aclent_t *aclbufp = get_acl(pathp, &aclcnt);
	//int aclcnt;
	acl_t *aclp;

	aclp = acl_get_file(pathp,
		((what & ACL_PRINT_ACCESS_ACL)  ? ACL_TYPE_ACCESS  : 0) |
		((what & ACL_PRINT_DEFAULT_ACL) ? ACL_TYPE_DEFAULT : 0));
	if (aclp == NULL)
		goto fail;

	acl_find(aclp, &user_obj, &group_obj, NULL, NULL);
	if (!(user_obj && group_obj)) {
		errno = EINVAL;
		goto fail;
	}

	if (opt_acl_only && acl_equiv_mode(aclp, NULL) == 0) {
		return 0;
	}

	if (*pathp == '/' && opt_strip_leading_slash) {
		if (!absolute_warning) {
			fprintf(stderr, "%s: removing leading '/' "
					"from absolute path names\n",
				progname);
			absolute_warning = 1;
		}

		if (*(pathp+1) == '\0')
			prefix = ".";
		pathp++;
	}

	/*
	if (*pathp == '.' && *(pathp+1) == '/')
		pathp += 2;
	*/

	if (!opt_no_comments) {
		printf("# file: %s%s\n", prefix, pathp);
		if ((owner = user_name(user_obj->a_id)) != NULL)
			printf("# owner: %s\n", owner);
		else
			printf("# owner: %d\n", (int)user_obj->a_id);
		if ((group = group_name(group_obj->a_id)) != NULL)
			printf("# group: %s\n", group);
		else
			printf("# group: %d\n", (int)group_obj->a_id);
	}

	if (acl_print(stdout, aclp, what |
				    ACL_PRINT_ACCESS_ACL |
				    ACL_PRINT_DEFAULT_ACL) == -1)
		goto fail;
	printf("\n");

	acl_free(aclp);
	return 0;

fail:
	if (aclp != NULL)
		acl_free(aclp);
	return -1;
}


void help(void)
{
	printf("get file access control lists\n"
		"usage: %s " CMD_LINE_SPEC "\n"
		"\t-x\tdisplay the file access control list\n"
		"\t-d\tdisplay the default access control list\n"
		"\t-c\tdo not display comment header\n"
		"\t-e\tprint all (-E = no) effective rights\n"
		"\t-s\tskip files that only have the base entries\n"
		"\t-a\tinclude .dot files\n"
		"\t-R\trecurse into subdirectories\n"
		"\t  \t(-Y = follow symlinks, -5 = postorder)\n"
		"\t-P\tdon't strip leading '/'\n"
		"\t-v\tprint version and exit\n"
		"\t-h\thelp\n",
		progname);
}


char *next_line(FILE *file)
{
	static char line[_POSIX_PATH_MAX], *c;
	if (!fgets(line, sizeof(line), file))
		return NULL;

	c = strrchr(line, '\0');
	while (c > line && (*(c-1) == '\n' ||
			   *(c-1) == '\r')) {
		c--;
		*c = '\0';
	}
	return line;
}


int main(int argc, char *argv[])
{
	int opt;
	int what = ACL_PRINT_EFFECTIVE;
	char *line;

	progname = argv[0];

	if (getenv("POSIXLY_CORRECT"))
		posixly_correct = 1;

	while ((opt = getopt(argc, argv, CMD_LINE_OPTIONS)) != -1) {
		switch (opt) {
			case 'a':  /* acl only */
				what |= ACL_PRINT_ACCESS_ACL;
				break;

			case 'd':  /* default acl only */
				what |= ACL_PRINT_DEFAULT_ACL;
				break;

			case 'c':  /* no comments */
				opt_no_comments = 1;
				break;

			case 'e':  /* all #effective comments */
				what |= ACL_PRINT_ALL_EFFECTIVE;
				break;

			case 'E':  /* no #effective comments */
				what &= ~(ACL_PRINT_EFFECTIVE |
				          ACL_PRINT_ALL_EFFECTIVE);
				break;

			case 'R':  /* recursive */
				opt_recurse = 1;
				break;

			case 'Y':  /* recursive, follow symlinks */
				opt_recurse = 1;
				opt_follow_symlinks = 1;
				break;

			case '5':  /* recursive, post-order */
				opt_recurse = 1;
				opt_postorder = 1;
				break;

			case 'P':  /* don't strip leading slash */
				opt_strip_leading_slash = 0;
				break;

			case 's':  /* disregard files without extended ACL */
				opt_acl_only = 1;
				break;

			/*case 'b':
				opt_oneline = 1;
				break;*/

			case 'v':  /* print version */
				printf("%s " VERSION "\n", progname);
				return 0;

			case 'h':  /* help */
				help();
				return 0;

			case ':':  /* option missing */
			case '?':  /* unknown option */
			default:
				goto synopsis;
		}
	}

	if (!(what & (ACL_PRINT_ACCESS_ACL |
	              ACL_PRINT_DEFAULT_ACL))) {
		if (posixly_correct)
			what |= ACL_PRINT_ACCESS_ACL;
		else
			what |= (ACL_PRINT_ACCESS_ACL |
			         ACL_PRINT_DEFAULT_ACL);
	}
		
	if ((optind == argc) /*&& !posixly_correct*/)
		goto synopsis;

	do {
		if (optind == argc ||
		    strcmp(argv[optind], "-") == 0) {
			while ((line = next_line(stdin)) != NULL) {
				if (*line == '\0')
					goto synopsis;

				if (do_get(line, what, 1) == -1) {
					fprintf(stderr, "%s: %s: %s\n",
					        progname, line,
						strerror(errno));
					had_errors++;
				}
			}
			if (ferror(stdin)) {
				fprintf(stderr, "%s: standard input: %s\n",
				        progname, strerror(errno));
				had_errors++;
			}
		} else {
			if (do_get(argv[optind], what, 1) == -1) {
				fprintf(stderr, "%s: %s: %s\n", progname,
					argv[optind], strerror(errno));
				had_errors++;
			}
		}
		optind++;
	} while (optind < argc);

	return had_errors ? 1 : 0;

synopsis:
	fprintf(stderr, "usage: %s " CMD_LINE_SPEC "\n"
	       "Try `%s -h' for more information.\n",
				 progname, progname);
	return 2;
}

