/*
  File: progs/setfacl.c
  (Linux Access Control List Management)

  Last modified: Wed 22 Sep 1999, 23:43

  Copyright (C) 1999 by Andreas Gruenbacher
  <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/acl.h>
#include <dirent.h>
#include "acl_util.h"


#define VERSION "0.5.0"

/*
	'-' stands for `process non-option arguments in loop'
*/
#define CMD_LINE_OPTIONS "-bknds:S:m:M:x:X:vhtRY5B:"
#define CMD_LINE_SPEC "[-bkndtRY5vh] [-B file] " \
                      "{ -s|-S|-m|-M|-x|-X ... } ... file ..."


const char *progname;
int opt_recurse = 0;  /* recurse */
int opt_follow_symlinks = 0;  /* follow symlinks */
int opt_recalculate = 1;  /* recalculate mask entry */
int opt_promote = 0;  /* promote access ACL to default ACL */
int opt_test = 0;  /* do not write to the file system.
                      Print what would happen instead. */
int opt_postorder = 0;  /* traverse in postorder (default is preorder) */
int had_errors = 0, had_warnings = 0;
int chown_error = 0;
int promote_warning = 0;


acl_entry_t *
find_entry(
	acl_t *aclp,
	int type,
	uid_t id)
{
	acl_entry_t *aclentp = NULL;

	while ((aclentp = acl_iterate(aclp, aclentp)) != NULL) {
		if ((aclentp->a_type == type) &&
		    ((((type & ~ACL_TYPE_DEFAULT) != ACL_USER &&
		      (type & ~ACL_TYPE_DEFAULT) != ACL_GROUP) ||
		      id == ACL_UNDEFINED_ID) ||
		     aclentp->a_id == id))
			return aclentp;
	}
	return NULL;
}


int
clone_entry(
	acl_t *aclp,
	int type,
	int newtype)
{
	acl_entry_t *ap, *ap2;
	ap = find_entry(aclp, type, ACL_UNDEFINED_ID);
	if (ap) {
		if (acl_create_entry(aclp, &ap2) != 0) {
			errno = ENOMEM;
			return -1;
		}
		acl_copy_entry(ap2, ap);
		ap2->a_type = newtype;
	} else {
		return 1;
	}
	return 0;
}


void
print_test(
	FILE *file,
	const char *pathp,
	const acl_t *aclp)
{
	acl_entry_t *ap;
	uid_t uid = ACL_UNDEFINED_ID;
	gid_t gid = ACL_UNDEFINED_ID;
	const char *owner, *group;

	ap = find_entry((acl_t *)aclp, ACL_USER_OBJ, ACL_UNDEFINED_ID);
	if (ap)
		uid = ap->a_id;
	ap = find_entry((acl_t *)aclp, ACL_GROUP_OBJ, ACL_UNDEFINED_ID);
	if (ap)
		gid = ap->a_id;

	fprintf(file, "# file: %s\n", pathp);
	if (uid != ACL_UNDEFINED_ID) {
		if ((owner = user_name(uid)) != NULL)
			fprintf(file, "# owner: %s\n", owner);
		else
			fprintf(file, "# owner: %d\n", (int)uid);
	}
	if (gid != ACL_UNDEFINED_ID) {
		if ((group = group_name(gid)) != NULL)
			fprintf(file, "# group: %s\n", group);
		else
			fprintf(file, "# group: %d\n", (int)gid);
	}
	acl_print(file, aclp, ACL_PRINT_ACCESS_ACL |
				ACL_PRINT_DEFAULT_ACL |
				ACL_PRINT_EFFECTIVE);
	fprintf(file, "\n");
}


acl_t *modify_acl(const acl_t *aclp, const acl_t *aclcmdp)
{
	acl_t *resp = acl_dup(aclp);
	acl_entry_t *entp = NULL, *ap;
	int cmd, type;
	int access_acl_entries, default_acl_entries;
	int access_acl_modified = 0, default_acl_modified = 0;
	int access_mask_given = 0, default_mask_given = 0;
	mode_t access_mask, default_mask;

	if (resp == NULL) {
		errno = ENOMEM;
		return NULL;
	}

	while ((entp = acl_iterate((acl_t *)aclcmdp, entp))) {
		if (opt_promote) {
			if (entp->a_type & ACL_TYPE_DEFAULT) {
				if (!promote_warning) {
					fprintf(stderr, "%s: ignoring default "
					        "ACL entries in input while "
						"in -d mode\n",
						progname);
					had_warnings++;
					promote_warning = 1;
				}
				continue;
			}
			entp->a_type |= ACL_TYPE_DEFAULT;
		}

		cmd = entp->a_type  & ACL_SEQ_CMD;
		type = entp->a_type & ~ACL_SEQ_CMD;

		if (type & ACL_TYPE_DEFAULT)
			default_acl_modified = 1;
		else
			access_acl_modified = 1;

		if (type == ACL_MASK_OBJ)
			access_mask_given = 1;
		if (type == ACL_DEF_MASK_OBJ)
			default_mask_given = 1;

		switch(cmd) {
			case ACL_SEQ_REMOVE_ENTRY:
				ap = find_entry(resp, type, entp->a_id);
				if (ap)
					acl_delete_entry(resp, ap);
				break;

			case ACL_SEQ_SUBTRACT:
				ap = find_entry(resp, type, entp->a_id);
				if (ap) {
					ap->a_perm &= ~entp->a_perm;
				}
				break;

			case ACL_SEQ_ADD:
				ap = find_entry(resp, type, entp->a_id);
				if (ap) {
					ap->a_perm |= entp->a_perm;
				} else {
					if (acl_create_entry(resp, &ap) != 0) {
						errno = ENOMEM;
						goto fail;
					}
					ap->a_type = type;
					ap->a_id = entp->a_id;
					ap->a_perm = entp->a_perm;
				}
				break;

			case ACL_SEQ_REPLACE:
				ap = find_entry(resp, type, entp->a_id);
				if (ap) {
					ap->a_perm = entp->a_perm;
				} else {
					if (acl_create_entry(resp, &ap) != 0) {
						errno = ENOMEM;
						goto fail;
					}
					ap->a_type = type;
					ap->a_id = entp->a_id;
					ap->a_perm = entp->a_perm;
				}
				break;

			case ACL_SEQ_REMOVE_DEFAULT:
				access_acl_modified = 1;
				default_acl_modified = 1;
			restart_remove_default:
				ap = NULL;
				while ((ap = acl_iterate(resp, ap))) {
					if (ap->a_type & ACL_TYPE_DEFAULT) {
						acl_delete_entry(resp, ap);
						goto restart_remove_default;
					}
				}
				break;

			case ACL_SEQ_REMOVE_EXTENDED:
				access_acl_modified = 1;
				default_acl_modified = 1;
			restart_remove_all:
				ap = NULL;
				while ((ap = acl_iterate(resp, ap))) {
					 switch(ap->a_type) {
					 	case ACL_USER_OBJ:
						case ACL_GROUP_OBJ:
						case ACL_OTHER_OBJ:
							continue;
					}
					acl_delete_entry(resp, ap);
					goto restart_remove_all;
				}
				break;

			case ACL_SEQ_REMOVE_ALL:
				access_acl_modified = 1;
				default_acl_modified = 1;
				acl_free(resp);
				resp = acl_init(1);
				if (resp == NULL) {
					errno = ENOMEM;
					goto fail;
				}
				break;

			default:
				errno = EINVAL;
				goto fail;
		}
	}

	/* Check what's in the ACL */
	access_acl_entries = 0;
	default_acl_entries = 0;
	ap = NULL;
	while((ap = acl_iterate((acl_t *)resp, ap))) {
		type = ap->a_type & ~ACL_SEQ_CMD;
		if (type & ACL_TYPE_DEFAULT)
			default_acl_entries |= (type & ~ACL_TYPE_DEFAULT);
		else
			access_acl_entries |= type;
	}

	/* Try to fill in missing entries */
	errno = EINVAL;
	if ((access_acl_entries & (ACL_USER | ACL_GROUP)) &&
	    !(access_acl_entries & ACL_MASK_OBJ)) {
		if (clone_entry(resp, ACL_GROUP_OBJ, ACL_MASK_OBJ) == 0)
			access_acl_entries |= ACL_MASK_OBJ;
	}
	if (default_acl_entries && !(default_acl_entries & ACL_USER_OBJ)) {
		if (clone_entry(resp, ACL_USER_OBJ, ACL_DEF_USER_OBJ) == 0)
			default_acl_entries |= ACL_USER_OBJ;
	}
	if (default_acl_entries && !(default_acl_entries & ACL_GROUP_OBJ)) {
		if (clone_entry(resp, ACL_GROUP_OBJ, ACL_DEF_GROUP_OBJ) == 0)
			default_acl_entries |= ACL_GROUP_OBJ;
	}
	if ((default_acl_entries & (ACL_USER | ACL_GROUP)) &&
	    !(default_acl_entries & ACL_MASK_OBJ)) {
		if (clone_entry(resp, ACL_DEF_GROUP_OBJ, ACL_DEF_MASK_OBJ) == 0)
			default_acl_entries |= ACL_MASK_OBJ;
	}
	if (default_acl_entries && !(default_acl_entries & ACL_OTHER_OBJ)) {
		if (clone_entry(resp, ACL_OTHER_OBJ, ACL_DEF_OTHER_OBJ) == 0)
			default_acl_entries |= ACL_OTHER_OBJ;
	}

	if (opt_recalculate) {
		access_mask = 0;
		default_mask = 0;
		ap = NULL;
		while ((ap = acl_iterate(resp, ap)) != NULL) {
			switch(ap->a_type) {
				case ACL_USER:
				case ACL_GROUP_OBJ:
				case ACL_GROUP:
					access_mask |= ap->a_perm;
					break;
				case ACL_DEF_USER:
				case ACL_DEF_GROUP_OBJ:
				case ACL_DEF_GROUP:
					default_mask |= ap->a_perm;
					break;
			}
		}

		ap = find_entry(resp, ACL_MASK_OBJ, ACL_UNDEFINED_ID);
		if (ap && access_acl_modified && !access_mask_given)
			ap->a_perm = access_mask;

		ap = find_entry(resp, ACL_DEF_MASK_OBJ, ACL_UNDEFINED_ID);
		if (ap && default_acl_modified && !default_mask_given)
			ap->a_perm = default_mask;
	}

	return resp;

fail:
	if (resp)
		acl_free(resp);
	return NULL;
}


int set_operation(const char *pathp, const acl_t *aclcmdp)
{
	acl_t *aclp = NULL, *acl2p = NULL;
	int error, which_entry;
	char *acltext;

	aclp = acl_get_file(pathp, ACL_TYPE_ACCESS |
	                           ACL_TYPE_DEFAULT);
	if (aclp == NULL)
		return -1;

	acl2p = modify_acl(aclp, aclcmdp);
	if (acl2p == NULL)
		goto fail;

	/* check the resulting ACL */
	error = acl_check(acl2p, &which_entry);
	if (error < 0) {
		acltext = acl_to_text(acl2p);
		fprintf(stderr, "%s: resulting ACL `%s', entry %d: %s\n",
			progname, acltext, which_entry,
			acl_strerror(-error));
		free(acltext);
		had_errors++;
	} else {
		if (opt_test) {
			print_test(stdout, pathp, acl2p);
		} else {
			if (acl(pathp, ACL_SETACL, acl2p->entries,
			                                    acl2p->entp) != 0) {
				goto fail;
			}
		}
	}

	if (aclp)
		acl_free(aclp);
	if (acl2p)
		free(acl2p);

	return 0;

fail:
	if (aclp)
		acl_free(aclp);
	if (acl2p)
		free(acl2p);
	return -1;
}


int do_set(const char *pathp, acl_t *aclcmdp, int toplevel)
{
	struct stat buf;
	DIR *dir;
	struct dirent *dirent;
	char *ipathp = (char*)pathp;
	char *subpathp = NULL;
	char *lastslash;
	int recurse = 0;
	int skip = 0;

	if (lstat(pathp, &buf) != 0)
		goto fail;

	lastslash = strrchr(pathp, '/');

	switch (buf.st_mode & S_IFMT) {
		case S_IFLNK:  /* symbolic link */
			skip = (!toplevel || opt_follow_symlinks);
			/* fall through */

		case S_IFDIR: /* directory */
			/* make sure it ends in a slash... */
			if (!lastslash || *(lastslash+1) != '\0') {
				ipathp = (char*)malloc(strlen(pathp)+2);
				if (ipathp == NULL)
					{ errno = ENOMEM; goto fail; }
				strcpy(ipathp, pathp);
				strcat(ipathp, "/");
			}

			if ((buf.st_mode & S_IFMT) == S_IFDIR)
				recurse = opt_recurse;
			else
				recurse = opt_recurse && opt_follow_symlinks;
			break;

		case S_IFREG:  /* regular file */
		case S_IFBLK:  /* block device */
		case S_IFCHR:  /* character device */
		case S_IFIFO:  /* named pipe */
		case S_IFSOCK:  /* UNIX domain socket */
		/* make sure it doesn't end in a slash... */
		if (lastslash && (*lastslash+1) == '\0') {
			ipathp = (char*)malloc(strlen(pathp)+1);
			if (ipathp == NULL)
				{ errno = ENOMEM; goto fail; }
			strcpy(ipathp, pathp);
			lastslash = strrchr(ipathp, '/');
			assert (lastslash != NULL);
			*lastslash = '\0';
		}
		break;

		default:
			errno = EINVAL;
			goto fail;
	}

	if (!skip && !opt_postorder) {
		if (set_operation(pathp, aclcmdp) == -1) {
			if (recurse && errno == EACCES) {
				fprintf(stderr, "%s: %s: %s\n", progname,
				        pathp, strerror(errno));
				had_errors++;
				return 0;
			} else {
				/* error is handled outside of this function */
				goto fail;
			}
		}
	}

	if (recurse) {
		dir = opendir(ipathp);
		if (dir == NULL) {
			if (errno == EACCES) {
				fprintf(stderr, "%s: %s\n",
				        pathp, strerror(errno));
				had_errors++;
				return 0;
			} else {
				goto fail;
			}
		}
		while ((dirent = readdir(dir)) != NULL) {
			if (strcmp(dirent->d_name, ".")  != 0 &&
			    strcmp(dirent->d_name, "..") != 0) {
				subpathp = (char*)malloc(strlen(ipathp) +
				                    strlen(dirent->d_name) + 1);
				if (subpathp == NULL)
					goto fail;

				strcpy(subpathp, ipathp);
				strcat(subpathp, dirent->d_name);
				do_set(subpathp, aclcmdp, 0); /* recurse */
				free(subpathp);
			}
		}
		closedir(dir);
	}

	if (!skip && opt_postorder) {
		if (set_operation(pathp, aclcmdp) == -1) {
			if (recurse && errno == EACCES) {
				fprintf(stderr, "%s: %s: %s\n", progname,
				        pathp, strerror(errno));
				had_errors++;
				return 0;
			} else {
				/* error is handled outside of this function */
				goto fail;
			}
		}
	}

	if (ipathp != pathp)
		free(ipathp);
	return 0;

fail:
	if (ipathp != pathp)
		free(ipathp);
	return -1;
}


int restore(FILE *file) {
	char *pathp = NULL;
	acl_t *aclp;
	acl_entry_t *ap;
	uid_t uid;
	gid_t gid;
	int line = 1, l;
	int status = 0;

	while ((aclp = acl_read(file, ACL_PARSE_MULTI |
	                              ACL_PARSE_WITH_PERM |
	                              ACL_PARSE_NO_RELATIVE, &l, NULL,
			                                     &pathp)) != NULL) {
		if (pathp == NULL) {
			fprintf(stderr, "%s: ACL description beginnig at line "
			               "%d: no file name found, exiting\n",
				       progname, line);
			status = 2;
			had_errors++;
			break;
		}

		line += l;

		if (opt_test) {
			print_test(stdout, pathp, aclp);
		} else {
			ap = find_entry(aclp, ACL_USER_OBJ, ACL_UNDEFINED_ID);
			if (ap)
				uid = ap->a_id;
			else
				uid = ACL_UNDEFINED_ID;
			ap = find_entry(aclp, ACL_GROUP_OBJ, ACL_UNDEFINED_ID);
			if (ap)
				gid = ap->a_id;
			else
				gid = ACL_UNDEFINED_ID;

			if ((uid != ACL_UNDEFINED_ID ||
			     gid != ACL_UNDEFINED_ID)) {
				if (geteuid() == 0) {
					if (chown(pathp, uid, gid) != 0) {
						fprintf(stderr, "%s: %s: "
						        "chown(): %s\n",
						        progname, pathp,
							strerror(errno));
						status = 1;
						had_errors++;
					}
				} else if (!chown_error) {
					fprintf(stderr, "%s: only root can "
					        "chown()\n", progname);
					status = 1;
					chown_error = 1;
					had_errors++;
				}
			}

			if (acl(pathp, ACL_SETACL, aclp->entries,
			                                     aclp->entp) != 0) {
				fprintf(stderr, "%s: %s: acl(): %s\n",
				                progname, pathp,
						strerror(errno));
				status = 1;
				had_errors++;
			}
		}
	}

	if (errno != 0 || ferror(file)) {
		fprintf(stderr, "%s: restore: %s\n",
		        progname, strerror(errno));
		status = 1;
		had_errors++;
	}

	return status;
}


void help()
{
	printf("set file access control lists\n"
		"usage: %s " CMD_LINE_SPEC  "\n"
		"\t-s acl file ...  /  -S acl_file file ...\n"
		"\t\tset ACL of file(s), replace current ACL\n\n"
		"\t-m acl file ...  /  -M acl_file file ...\n"
		"\t\tmodify current ACL of file(s)\n\n"
		"\t-x acl file ...  /  -X acl_file ...\n"
		"\t\tdelete entries in ACL of file(s)\n\n"
		"\t-b\tremove all extended ACL entries\n"
		"\t-k\tremove default ACL\n"
		"\t-n\tdon't recalculate effective rights mask\n"
		"\t-d\toperations apply to the default ACL\n"
		"\t-R\trecurse into subdirectories\n"
		"\t  \t(-Y = follow symlinks, -5 = post-order)\n"
		"\t-t\ttest mode (no changes to files)\n"
		"\t-B\trestore acls (inverse of `getfacl -R')\n"
		"\t-v\tprint version and exit\n"
		"\t-h\thelp\n"
		/*"\t-i\tfurther explain syntax of options\n"*/,
		progname);
}


#if 0
void explain()
{
	printf("\n"
		"acl_perms -- a comma-separated list of ACL entries + permissons.\n"
		"acl_ents  -- a comma-separated list of ACL entries, identical\n"
		"             to acl_perms but without mode specification.\n"
		"acl_file  -- file containing an ACL. The contents are similar to\n"
		"             what getfacl prints. Read from stdin if acl_file='-'.\n"
		"\n"
		"ACL entries:\n"
		"\n"
		"[default:]user:[<id>]:<mode>  --  user (owner) permissions\n"
		"[default:]group:[<id>]:<mode> --  group (owning group) permissions\n"
		"[default:]other:<mode>        --  other users permissions\n"
		"[default:]mask:<mode>         --  maximum permissions mask\n"
		"\n"
		"default: -- default permissions for directories that are inherited\n"
		"            by files and subdirectories created in that directory.\n"
		"<id>     -- user / group name or number. If omitted, the ACL entry\n"
		"            applies to the owner / owning group.\n"
		"\n"
		"d[efault], u[ser], g[roup] and o[ther] can be abbreviated.\n"
		"<mode> is any one of 'r', 'w', 'x', optionally padded with '-'.\n"
		"\n"
		"The mask entry limits the effective rights of all users and\n"
		"groups, except the file owner and `other's.\n\n");
}
#endif


int add_seq_operation(acl_t *aclcmdp, int seq_op) {
	acl_entry_t *ap;
	int error;

	error = acl_create_entry(aclcmdp, &ap);
	if (error != 0) {
		return error;
	}
	ap->a_type = seq_op;
	return 0;
}


char *next_line(FILE *file)
{
	static char line[_POSIX_PATH_MAX], *c;
	if (!fgets(line, sizeof(line), file))
		return NULL;

	c = strrchr(line, '\0');
	while (c > line && (*(c-1) == '\n' ||
			   *(c-1) == '\r')) {
		c--;
		*c = '\0';
	}
	return line;
}


int main(int argc, char *argv[])
{
	int opt;
	int saw_files = 0;
	int status = 0;
	FILE *file;
	const char *which;
	int lineno;
	int error;
	acl_t *aclp = NULL, *aclcmdp = NULL;
	acl_entry_t *ap1, *ap2;
	int seq_cmd, parse_mode;
	char *line;
	
	progname = argv[0];

	if ((aclcmdp = acl_init(1)) == NULL) {
		errno = ENOMEM;
		fprintf(stderr, "%s: %s\n", progname, strerror(errno));
	}

	while ((opt = getopt(argc, argv, CMD_LINE_OPTIONS)) != -1) {
		switch (opt) {
			case 'b':  /* remove all extended entries */
			case 'k':  /* remove default ACL */
				if (!add_seq_operation(aclcmdp,
				  (opt == 'b') ? ACL_SEQ_REMOVE_EXTENDED
				               : ACL_SEQ_REMOVE_DEFAULT) == 0) {
					fprintf(stderr, "%s: %s\n",
					        progname, strerror(errno));
					status = 2;
					goto cleanup;
				}
				break;


			case 'n':  /* do not recalculate mask */
				opt_recalculate = 0;
				break;

			case 'd':  /*  operations apply to default ACL */
				opt_promote = 1;
				break;

			case 's':  /* set */
				if (!add_seq_operation(aclcmdp,
				                ACL_SEQ_REMOVE_ALL) == 0) {
					fprintf(stderr, "%s: %s\n",
					        progname, strerror(errno));
					status = 2;
					goto cleanup;
				}

				seq_cmd = ACL_SEQ_REPLACE;
				parse_mode = ACL_PARSE_WITH_PERM |
				             ACL_PARSE_NO_RELATIVE;
				goto set_modify_delete;

			case 'm':  /* modify */
				seq_cmd = ACL_SEQ_REPLACE;
				parse_mode = ACL_PARSE_WITH_PERM |
				             ACL_PARSE_RELATIVE;
				goto set_modify_delete;

			case 'x':  /* delete */
				seq_cmd = ACL_SEQ_REMOVE_ENTRY;
				parse_mode = ACL_PARSE_NO_PERM |
				             ACL_PARSE_NO_RELATIVE;
				goto set_modify_delete;

			set_modify_delete:
				aclp = parse_acl(optarg, &which, parse_mode);

				if (!aclp) {
					if (*which == '\0') {
						fprintf(stderr, "%s: option "
						        "-%c incomplete\n",
							progname, opt);
					} else {
						fprintf(stderr, "%s: option "
						        "-%c: %s (near "
							"character %d)\n",
							progname, opt,
							strerror(errno),
							which - optarg + 1);
					}
					status = 2;
					goto cleanup;
				}

			do_set_modify_delete:
				if (saw_files) {
					acl_free(aclcmdp);
					aclcmdp = acl_init(1);
					if (aclcmdp == NULL) {
						status = 2;
						goto cleanup;
					}
					saw_files = 0;
				}

				ap1 = NULL;
				while ((ap1 = acl_iterate(aclp, ap1))) {
					ap1->a_type |= seq_cmd;
					error = acl_create_entry(aclcmdp, &ap2);
					if (error != 0) {
						fprintf(stderr, "%s: %s\n",
						        progname,
							strerror(errno));
						status = 2;
						goto cleanup;
					}
					*ap2 = *ap1;
				}
				acl_free(aclp);
				aclp = NULL;
				break;


			case 'S':  /* set from file */
				if (!add_seq_operation(aclcmdp,
				                ACL_SEQ_REMOVE_ALL) == 0) {
					fprintf(stderr, "%s: %s\n",
					        progname, strerror(errno));
					status = 2;
					goto cleanup;
				}

				seq_cmd = ACL_SEQ_REPLACE;
				parse_mode = ACL_PARSE_WITH_PERM |
				             ACL_PARSE_NO_RELATIVE;
				goto set_modify_delete_from_file;
				
			case 'M':  /* modify from file */
				seq_cmd = ACL_SEQ_REPLACE;
				parse_mode = ACL_PARSE_WITH_PERM |
				             ACL_PARSE_RELATIVE;
				goto set_modify_delete_from_file;

			case 'X':  /* delete from file */
				seq_cmd = ACL_SEQ_REMOVE_ENTRY;
				parse_mode = ACL_PARSE_NO_PERM |
				             ACL_PARSE_NO_RELATIVE;
				goto set_modify_delete_from_file;

			set_modify_delete_from_file:
				if (strcmp(optarg, "-") == 0) {
					file = stdin;
				} else {
					file = fopen(optarg, "r");
					if (file == NULL) {
						fprintf(stderr, "%s: %s: %s\n",
							progname, optarg,
							strerror(errno));
						status = 2;
						goto cleanup;
					}
				}

				aclp = acl_read(file, parse_mode, &lineno,
				                NULL, NULL);
				
				if (file != stdin) {
					fclose(file);
				}

				if (!aclp) {
					if (errno == 0)
						errno = EINVAL;

					fprintf(stderr, "%s: %s in line %d\n",
					        progname, strerror(errno),
						lineno);
					if (file != stdin) {
						fprintf(stderr, " of file %s\n",
						        optarg);
					}
					status = 2;
					goto cleanup;
				}
				goto do_set_modify_delete;


			case '\1':  /* file argument */
				if (aclcmdp->entries == 0)
					goto synopsis;
				saw_files = 1;

				if (strcmp(optarg, "-") == 0) {
					while ((line = next_line(stdin))) {
						if (do_set(line,
						           aclcmdp, 1) != 0) {
							fprintf(stderr,
							        "%s: %s: %s\n",
							        progname,
							        line,
							        strerror(errno)
							       );
							status = 1;
							goto cleanup;
						}
					}
				} else {
					if (do_set(optarg, aclcmdp, 1) != 0) {
						fprintf(stderr, "%s: %s: %s\n",
							progname, optarg,
							strerror(errno));
						status = 1;
						goto cleanup;
					}
				}
				break;


			case 'B':
				saw_files = 1;
				if (strcmp(optarg, "-") == 0) {
					file = stdin;
				} else {
					file = fopen(optarg, "r");
					if (file == NULL) {
						fprintf(stderr, "%s: %s: %s\n",
							progname, optarg,
							strerror(errno));
						status = 2;
						goto cleanup;
					}
				}

				status = restore(file);

				if (file != stdin) {
					fclose(file);
				}

				if (status != 0)
					goto cleanup;
				break;


			case 't':  /* test mode */
				opt_test = 1;
				break;

			case 'R':
				opt_recurse = 1;
				break;

			case 'Y':
				opt_recurse = 1;
				opt_follow_symlinks = 1;
				break;

			case '5':
				opt_recurse = 1;
				opt_postorder = 1;
				break;

			case 'v':
				printf("%s " VERSION "\n", progname);
				status = 0;
				goto cleanup;

			/*case 'i':
				explain();
				status = 0;
				goto cleanup;*/

			case 'h':
				help();
				status = 0;
				goto cleanup;

			case ':':  /* option missing */
			case '?':  /* unknown option */
			default:
				goto synopsis;
		}
	}

	if (!saw_files)
		goto synopsis;

	goto cleanup;

synopsis:

	fprintf(stderr, "usage: %s " CMD_LINE_SPEC "\n"
	       "Try `%s -h' for more information.\n",
				 progname, progname);
	status = 2;
	goto cleanup;

cleanup:

	if (aclcmdp)
		acl_free(aclcmdp);
	if (aclp)
		acl_free(aclp);
	return status;
}

